#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <getopt.h>
#include <assert.h>
#include <unistd.h>
#include <omp.h>

#include "quilting.h"

/** print a message and abort the execution */
#define FATAL(MSG)			\
do {				\
	fprintf(stderr, MSG "\n");	\
	abort();			\
	} while (0);

static int display_usage()
{
    printf("\nquilting_main -p patch_sz -r ratio -t tolerance -o overlap -s seed "
           "src_im.png out_im.png\n\n");
    printf("Required parameters:\n");
    printf("   src_im.png      		  :   name of the input PNG image\n");
    printf("   out_im.png		      :   name of the output PNG image\n\n");
    printf("   position_map.png       :   name of the position map PNG image\n");
    printf("   synthesis_map.png      :   name of the synthesis map PNG image\n\n");
    printf("Optionnal parameters:\n");
    printf("   patch_sz        :   int > size(src_im.png) that "
           "specifies the size of the patch used to synthesize "
           "out_im.png\n");
    printf("   ratio           :   float >= 1 that specifies the increase "
           "of size of out_im.png relatively to the size "
           "of src_im.png\n");
    printf("   tolerance       :   float > 0 to specify the error "
           "tolerance when comparing patches in src_im.png\n");
    printf("   overlap         :   0 < float < 1 to specify the patch overlap ratio\n");
    printf("   seed            :   unsigned int to specify the seed "
           "for the random number generator "
           "(seed = time(NULL) by default)\n");

    return (1);
}

int main(int argc, char *argv[])
{       
    /* 
     * Getting parameters: patch size, overlap size, number 
     * of blocks per row, number of blocks per column, tolerance 
     * for error
     */

    char *fname_in;         	/* input file name */
    char *fname_out;        	/* output file name */
	char *fname_positionMap;	/* position map file name */
	char *fname_synthesisMap;	/* synthesis map file name */

    Image src_im;           	/* input data */   
    Image out_im;           	/* output data */
    int patch_sz;           	/* patch size */
    float ratio;            	/* increase ratio for the
                             	 * output size (must be >1) */
    int overlap_sz;         	/* overlap size between blocks */
    int blocks_row;         	/* number of blocks per row*/ 
    int blocks_col;         	/* number of blocks per column*/
    
    unsigned long seed = 0; 	/* seed for the random number
                             	 * generator */
    int p_flag = 0;
    int s_flag = 0;
    int c;                  	/* getopt flag */    
    
    /* "default" value initialization */
    patch_sz = 40;
    ratio = 1.;
    float tolerance = .1;
    float overlap_ratio = .25;
    
    /* process the options and parameters */
    
    while ((c = getopt(argc, argv, "p:r:t:o:s:h")) != -1) {
        switch (c) {
        case 'p':
            /* patch size specified */
            p_flag = 1;
            patch_sz = atof(optarg);
            break;
        case 'r':
            /* increase ratio specified */
            //r_flag = 1;
            ratio = atof(optarg);
            if (ratio < 1)
                FATAL("Option usage: The ratio "
                      "specified by -r must be greater than 1.");
            break;
        case 't':
            /* error tolerance specified */
            //t_flag = 1;
            tolerance = atof(optarg);
            if (tolerance == 0.)
                FATAL("Option usage: The tolerance error  "
                      "specified by -t must be greater than 0.");
            break;
        case 'o':
            /* overlap ratio specified */
            overlap_ratio = atof(optarg);
            if (overlap_ratio <= 0. || overlap_ratio >= 1.)
                FATAL("Option usage: The overlap ratio  "
                      "specified by -o must be between 0. and 1.");
            break;
        case 's':
            /* seed specified */
            s_flag = 1;
            seed = (unsigned long) atol(optarg);
            break;
        case 'h':
            /* display usage */
            display_usage();
            return (-1);
        default:
            abort();
        }
    }
    
    /* process the non-option parameters */
    if (2 > (argc - optind)) {
        printf("The input and output image file names are missing\n\n");
        display_usage();
        return (-1);
    }
    fname_in = argv[optind++];
    fname_out = argv[optind++];
	fname_positionMap = argv[optind++];
    fname_synthesisMap = argv[optind++];
    
    /* Reading input image */
    src_im.img = io_png_read_f32(fname_in, &src_im.w, &src_im.h, &src_im.c);

    /* Initilization of the random number generator */
    if (s_flag == 0) {
		srand(time(NULL));
    }
    else {
		srand(seed);
    }

    /* deal with patch size, if it's bigger than the minimum lentgh of the input image abort*/
    if((p_flag==1) && ((patch_sz > (int) src_im.h) || (patch_sz > (int) src_im.w)))
    {
        printf("Conflict: the patch size should be smaller "
               "than the minimal length of the input image\n\n");
        display_usage();
        return (-1);
    }
    
    /* deal with patch size, if it's bigger than the minimum lentgh of the input image abort*/
    if((p_flag==1) && (patch_sz==0))
    {
        printf("Conflict: the patch size should be greater than zero\n\n");
        display_usage();
        return (-1);
    }

    /* compute overlap size */
    if ((overlap_sz = (int) round(patch_sz*overlap_ratio))==0 ||
         overlap_sz >= patch_sz )
		FATAL("Conflict: The overlap size should at least one pixel and less than the patch size");
    
    /* compute number of patches per row and per column needed to synthesize an (factor*M) x (factor*N) modulo (patch_sz) image */
    int M = roundf(ratio*src_im.h);
    int N = roundf(ratio*src_im.w);
    int C = src_im.c;
    float a;    
    if (src_im.w < src_im.h)
    {a = M - overlap_sz;}
    else
    {a = N - overlap_sz;}
    float b;
    if (src_im.h < src_im.w)
    {b = N - overlap_sz;}
    else 
    {b = M - overlap_sz;}
    float d = patch_sz - overlap_sz;
    blocks_row = (int) ceil(b/d);
    blocks_col = (int) ceil(a/d);

	Image position_map;// each pixel from in the source image is associated to a different color from a continuous colormap
	Image synth_map;// show for each synthesized patch its position in the source image

	//* Create position map*//
	initialize_image(&position_map, src_im.h, src_im.w, 3);
	create_position_map(position_map);

    //* Image synthesis *//
	out_im = quilting_synthesis(src_im, patch_sz, overlap_sz, blocks_row, blocks_col, tolerance, &position_map, &synth_map);
    
	//* Crop synthesis map *//
    Image synth_map_crop;
    if (crop_image(&synth_map_crop, &synth_map, M, N, 3))
    {
		FATAL("The size of the cropped image is bigger "
                      "than the image to crop");
    }

    //* Crop output image *//
    Image out_im_crop;
    if (crop_image(&out_im_crop, &out_im, M, N, C))
    {
        FATAL("The size of the cropped image is bigger "
                      "than the image to crop");
    }

	//* Writing position map in png file *//
	io_png_write_f32(fname_positionMap, position_map.img, position_map.w, position_map.h, position_map.c);
    
	//* Writing synthesis map in png file *//
    io_png_write_f32(fname_synthesisMap, synth_map_crop.img, synth_map_crop.w, synth_map_crop.h, synth_map_crop.c);

    /* Writing output image in png file */
    io_png_write_f32(fname_out, out_im_crop.img, out_im_crop.w, out_im_crop.h, out_im_crop.c);
                                              
    /* Release memory */
	free(position_map.img);
	free(synth_map.img);
	free(synth_map_crop.img);
    free(src_im.img);
    free(out_im.img);
	free(out_im_crop.img);
	
    return 0;
}
